(*                                                                          *)
(* (c) 2004, 2005 Anastasia Gornostaeva. <ermine@ermine.pp.ru>              *)
(*                                                                          *)

open Xml

let regexp bom = 0xFEFF
(* let regexp bom = 239 187 191 *)

let regexp space = ' ' | '\n' | '\t' | '\r'

let regexp any = xml_letter | xml_digit | xml_extender | xml_base_char 
               | xml_ideographic | xml_combining_char | xml_blank

let regexp name_char = xml_letter | xml_digit | '.' | '-' | '_' | ':' 
                     | xml_combining_char | xml_extender
let regexp name = (xml_letter | '_' | ':') (name_char*)

let regexp ent = "&lt;" | "&gt;" | "&amp;" | "&quot;" | "&apos;"

exception XmlError of string * string

let enclosed lexbuf offset_begin offset_end =
   let len = Ulexing.lexeme_length lexbuf in
      Ulexing.utf8_sub_lexeme lexbuf offset_begin
         (len - (offset_end + offset_begin))

let parse_error buf str =
   raise (XmlError (buf, str))

let rec token = lexer
   | bom? "<?xml" ->
	let _ = xmldeclattrs [] lexbuf in
	   token lexbuf
   | "<?xml-" ->
	let _ = pi lexbuf in
	   token lexbuf
   | "<?" ->
        let _ = pi lexbuf in
           token lexbuf

   | "<!--" -> 
	comment lexbuf;
	token lexbuf
   | "<!DOCTYPE" ->
	doctype lexbuf;
	token lexbuf
   | '<' ->
        let tag = name lexbuf in
        let empty, attrs = attributes [] lexbuf in
        let subels = match empty with
           | true -> 
		[]
           | false -> 
		elements tag [] lexbuf
        in
           Xmlelement (tag, attrs, subels)
   | space+ -> token lexbuf
   | _ -> 
	parse_error (Ulexing.utf8_lexeme lexbuf) "Invalid XML token"

and pi = lexer
   | "?>" -> ""
   | _ -> pi lexbuf

and xmldeclattrs acc = lexer
   | space* "?>" space* ->
	acc
   | space+ ->
	let attr = attribute lexbuf in
	   xmldeclattrs (attr :: acc) lexbuf
   | _ -> 
	parse_error (Ulexing.utf8_lexeme lexbuf) "Invalid XML [xmldecl]"

and comment = lexer
   | [^'-']* -> comment lexbuf
   | "-->" -> ""
   | "--" -> parse_error "--" "'--' not allowed inside comments"
   | "-" -> comment lexbuf
   | any -> parse_error (Ulexing.utf8_lexeme lexbuf) "Expecting comment end"

and doctype =  lexer
   | '>' -> ""
   | '[' [^"]"]* ']' -> doctype lexbuf  
   | _ -> doctype lexbuf


and element = lexer
   | name ->
	let tag = Ulexing.utf8_lexeme lexbuf in
	let empty, attrs = attributes [] lexbuf in
	   if empty then
	      Xmlelement (tag, attrs, [])
	   else
	      Xmlelement (tag, attrs, (elements tag [] lexbuf))
   | _ -> 
	parse_error (Ulexing.utf8_lexeme lexbuf) "Invalid XML [element]"

and elements tag acc = lexer
   | "<!--" -> 
	let _ = comment lexbuf in
	   elements tag acc lexbuf
   | "</" ->
	let endtag = name lexbuf in
	   if endtag = tag then 
	      begin
		 eat_end lexbuf;
		 List.rev acc
	      end
	   else
	      parse_error endtag ("Invalid XML end tag: expected " ^ tag)
   | '<' ->
	elements tag ((element lexbuf) :: acc) lexbuf
   | [^'<'] ->
	Ulexing.rollback lexbuf;
        let cdatatxt = cdata "" lexbuf in
	   elements tag ((Xmlcdata cdatatxt) :: acc) lexbuf
   | "<![CDATA[" ->
        let spec_cdata = special_cdata "" lexbuf in
           elements tag ((Xmlcdata spec_cdata) :: acc) lexbuf
   | any -> 
	parse_error (Ulexing.utf8_lexeme lexbuf) "Invalid XML [elements]"

and attributes attrs = lexer
   | space? "/>" -> 
	true, attrs
   | space? '>' -> 
	false, attrs
   | space+ ->
	let attr = attribute lexbuf in
           attributes (attr :: attrs) lexbuf
   | any ->
	parse_error (Ulexing.utf8_lexeme lexbuf) "No space here"

and name = lexer
   | name -> 
	Ulexing.utf8_lexeme lexbuf
   | _ -> 
	parse_error (Ulexing.utf8_lexeme lexbuf) "Invalid char in tag"

and attribute = lexer
   | name '=' ->
	let attrname = enclosed lexbuf 0 1 in
           attrname, attvalue lexbuf
   | _ -> 
	parse_error (Ulexing.utf8_lexeme lexbuf) "Invalid attribute name"

and attvalue = lexer
   | "\"" -> 
	attvalue_quot "" lexbuf
   | "'" -> 
	attvalue_apos "" lexbuf
   | _ -> 
	parse_error (Ulexing.utf8_lexeme lexbuf) "No quotes"

and attvalue_quot value = lexer
   | ([^ '<' '&' '\"' ] | ent)* ->
	attvalue_quot (value ^ (Ulexing.utf8_lexeme lexbuf)) lexbuf
   | "&#" ->
	attvalue_quot (value ^ unicode lexbuf) lexbuf
   | "\"" -> 
	value
   | any -> 
	parse_error (Ulexing.utf8_lexeme lexbuf) "Not expected here"

and attvalue_apos value = lexer
   | ([^ '<' '&' '\''] | ent)* ->
	attvalue_apos (value ^ (Ulexing.utf8_lexeme lexbuf)) lexbuf
   | "&#" ->
	attvalue_apos (value ^ unicode lexbuf) lexbuf
   | "'" -> 
	value
   | any -> 
	parse_error (Ulexing.utf8_lexeme lexbuf) "Not expected here"

and cdata acc = lexer
   | [^'<' '&']* ->
	cdata (acc ^ Ulexing.utf8_lexeme lexbuf) lexbuf
   | ent ->
	cdata (acc ^ Ulexing.utf8_lexeme lexbuf) lexbuf
   | "&#" ->
	cdata (acc ^ unicode lexbuf) lexbuf
   | '<' -> 
	Ulexing.rollback lexbuf;
	acc
   | any ->
	parse_error (Ulexing.utf8_lexeme lexbuf) "Invalid XML [cdata]"

and unicode = lexer
   | ['0'-'9']+ ->
	let i = Ulexing.utf8_lexeme lexbuf in
	   ((lexer 
	    | ';' -> ()
	    | any -> parse_error (Ulexing.utf8_lexeme lexbuf) "invalid unicode"
	    ) lexbuf);
           let b = Buffer.create 4 in
              Utf8.store b (int_of_string i);
	      Buffer.contents b
   | 'x' ['0'-'0' 'a'-'f' 'A'-'F']+ ->
	let i = "0" ^ Ulexing.utf8_lexeme lexbuf in
	   ((lexer 
	    | ';' -> ()
	    | any -> parse_error (Ulexing.utf8_lexeme lexbuf) "invalid unicode"
	    ) lexbuf);
           let b = Buffer.create 4 in
              Utf8.store b (int_of_string i);
	      Buffer.contents b
   | any ->
	parse_error (Ulexing.utf8_lexeme lexbuf) "invalid unicode"

and eat_end = lexer
   | space* '>' ->
	()
   | any ->
	parse_error (Ulexing.utf8_lexeme lexbuf) "Invalid XML [get_end]"

and special_cdata acc = lexer
(*
   | '<' -> special_cdata (acc ^ "&lt;") lexbuf
   | '>' -> special_cdata (acc ^ "&gt;") lexbuf
   | "'" -> special_cdata (acc ^ "&apos;") lexbuf
   | '\"' -> special_cdata (acc ^ "&quot;") lexbuf
   | "&" -> special_cdata (acc ^ "&amp;") lexbuf
*)
   | ent ->
	special_cdata (acc ^ Ulexing.utf8_lexeme lexbuf) lexbuf
   | "&#" ->
	special_cdata (acc ^ unicode lexbuf) lexbuf
   | "]]>" -> 
	acc
   | _ -> special_cdata (acc ^ (Ulexing.utf8_lexeme lexbuf)) lexbuf

let parse_string string =
   let lexbuf = Ulexing.from_utf8_string string in
      try 
	 token lexbuf
      with exn ->
	 raise exn

(*
let _ =
   let f = open_in "en.xml" in
   let b = Buffer.create 100 in
      (try
	 while true do
	    let str = input_line f in
	       Buffer.add_string b str;
	 done
      with _ -> ());
*)

